// alphas for TMDlib
#include "TMDlib.h"
#include <iostream>
#include <fstream>
#include <yaml-cpp/yaml.h>
#include <stdio.h>      /* printf, NULL */
#include <stdlib.h>     /* strtod */


using namespace TMDlib; 
using namespace std;

extern string pdfpath;

double TMD::get_key_val_as_double(string s)
/** Helper routine to return value from info file as double */
{
    double number;
    if ( s != "" ) {
       std::stringstream ss;
       ss << s;
       ss >> number;
       // cout << " get_key_val " << s << " number " << number << endl;
       return number; }
    else { return -9999.;}
}

int TMD::get_key_val_as_int(string s)
/** Helper routine to return value from info file as int */
{
    int number;
    // cout << " get_key_val " << s  << endl;
    if ( s != "" ) {
       std::stringstream ss;
       ss << s;
       ss >> number;
       // cout << " get_key_val " << s << " number " << number << endl;
       return number; }
    else { return -9999.;}
}

bool TMD::DoesFileExist (string name) {
/** Helper routine to check existence of file */
   // cout << " DoesFIle exit " << name << endl;
   if (FILE *file = fopen(name.c_str(), "r")) {
        fclose(file);
        return true;
    } else {
        return false;
    } 
     
}


void TMD::TMDinfo(const string name ) {
/** Helper to access inforamtion from info file using YAML.
   Old format (TMDlib1) is hard coded.
*/
    string pdfdir="" ;
    
//  here treat the TMDlib1 data sets   
    NewFormat = 0;  
    std::size_t found ;
    string setname = name ;
    found = name.find("GBW"); 
    if (found!=std::string::npos) {pdfdir="GBW/" ;}
    found = name.find("Bluem"); 
    if (found!=std::string::npos) {pdfdir="Blueml/" ;}
    found = name.find("ccfm"); 
    if (found!=std::string::npos) {pdfdir="CCFM/" ;}
    found= name.find("PB-NLO-HERAI+II");
    if (found!=std::string::npos) {pdfdir="PB-NLO-2018/";}
    found= name.find("PB-LO-HERAI+II");
    if (found!=std::string::npos) {pdfdir="PB-LO-2018/";}
    found= name.find("PB-NLO_ptoPb208");
    if (found!=std::string::npos) {pdfdir="PB-NLO_208_82/";}
    found= name.find("KS");
    if (found!=std::string::npos) {pdfdir="KS/";}
    found= name.find("Kutak");
    if (found!=std::string::npos) {pdfdir="KS/";}
    found= name.find("BHKS");
    if (found!=std::string::npos) {pdfdir="BHKS/";}
    found= name.find("EKMP");
    if (found!=std::string::npos) {pdfdir="EKMP/";}
    found= name.find("PB-EPPS16nlo_CT14nlo_Pb208");
    if (found!=std::string::npos) {pdfdir="EPPS16nlo_CT14nlo_Pb208/";}
    found= name.find("PB-nCTEQ15FullNuc_208_82");
    if (found!=std::string::npos) {pdfdir="nCTEQ15FullNuc_208_82/";}
    found= name.find("PB-gluon_D_c_ncteq1568CL_Pb_Nrep1000_hes");
    if (found!=std::string::npos) {pdfdir="gluon_D_c_ncteq1568CL_Pb_Nrep1000_hes/";}
    found= name.find("SBRS-2013-TMDPDFs");
    if (found!=std::string::npos) {pdfdir="SBRS/";}

    // checking for members
    int nmem = 0;
    const size_t slashpos = name.find("/");
    string smem;
    TMD_Mem="0000";
    setname = trim(name.substr(0, slashpos));
    if (slashpos != string::npos) {
       smem = name.substr(slashpos+1);
       cout << " smem " << smem << endl ;
       nmem = stoi( smem ) ;
       if( nmem < 10 )  { TMD_Mem = "000"+smem;}
       else if ( nmem < 100 )  { TMD_Mem = "00"+smem;}
       else if ( nmem < 1000 )  { TMD_Mem = "0"+smem;}
    }
    //else { cout << "No submembers  " << name<< " " << setname << endl;}

    
    if ( pdfdir == "" ) {
       cout << " pdfdir was not set: do we have a new set ? " << endl;
       
       cout << " name = " << setname << " with member " << nmem << endl;
       pdfdir=setname+"/";
       NewFormat = 1 ;
    }
    TMD_Name = setname ;
    TMD_Dir = pdfdir ;
    
    string filepath =pdfpath+pdfdir+setname+".info";
    
    
    
    if (!DoesFileExist(filepath)) { 
      cout << " TMDinfo: file does not exist " << filepath << endl;
      cout << " return on EXIT_FAILURE " << endl; 
      exit (EXIT_FAILURE);
      }
    
    if (filepath.empty()) std::cout << "Empty PDF file name given to Info::load  " << filepath << endl;
    // cout << " TMDinfo: " << filepath << " name = " << name << endl; 
    std::ifstream file(filepath);
    string docstr, line;
    while (getline(file, line)) {
      if (line == "---") break;
      docstr += line + "\n";
    }
    TMDLIB_YAML::Node doc = TMDLIB_YAML::Load(docstr);
    for (TMDLIB_YAML::const_iterator it = doc.begin(); it != doc.end(); ++it) {
      const string key = it->first.as<string>();
      const TMDLIB_YAML::Node& val = it->second;
      if (val.IsScalar()) {
         //  Scalar value
        // cout << key<< ": " <<val.as<string>() << endl;
         TMDdict[key] = val.as<string>();
      } else {
         // Process the sequence entries into a comma-separated string
         string seqstr = "";
         for (size_t i = 0; i < val.size(); ++i)
            seqstr += val[i].as<string>() + ((i < val.size()-1) ? "," : "");
         TMDdict[key] = seqstr;
        // cout << key << ":  " <<  seqstr << endl;
        }
    }
/*
    const string SearchKey = "SetDesc";
    std::cout << "SetDesc: " << TMDdict.find("SetDesc")->second << '\n';
    std::cout << "XMin: " << TMDdict.find("XMin")->second << '\n';
    std::cout << "XMax: " << TMDdict.find("XMax")->second << '\n';
    std::cout << "Flavors: " << TMDdict.find("Flavors")->second << '\n';
*/
     
}


string TMD::TMDgetDesc( ){
/** TMDlib helper to return Descripton from info file */

   return TMDdict.find("SetDesc")->second;
   

}
string TMD::TMDgetScheme( ){
/** TMDlib helper to return Scheme for TMD evolution from info file */

   return TMDdict.find("TMDScheme")->second;
   

}

double TMD::TMDgetLam4( ){
/** TMDlib helper to return QCDlam from info file */

   double QCDlam =  get_key_val_as_double(TMDdict.find("AlphaS_QCDlam")->second);
   if(QCDlam == 0 ) { 
      cout << " TMDgetLam4: QCDlam is not defined for iset = " << iset << endl; 
      return -9999.; }
   else { 
      return QCDlam;}
}

int TMD::TMDgetNf( ){
/** TMDlib helper to return Nr of flavors from info file */

     int index =  get_key_val_as_int(TMDdict.find("NumFlavors")->second);
    // cout << " TMDgetNF: " << TMDdict.find("NumFlavors")->second << " index = " << index <<  endl; 
     return index; 
   
}
int TMD::TMDgetNumMembers( ){
/** TMDlib helper to return Nr of Member from info file */

     int index =  get_key_val_as_int(TMDdict.find("NumMembers")->second);
    // cout << " TMDgetNumMembers: " << TMDdict.find("NumMembers")->second << " index = " << index <<  endl; 
     return index; 
   
}
int TMD::TMDgetOrderAlphaS( ){
/** TMDlib helper to return order of alphas (nr of loops) from info file */

     int index =  get_key_val_as_int(TMDdict.find("AlphaS_OrderQCD")->second);
     // cout << " TMDgetOrderAlphas: " << TMDdict.find("AlphaS_OrderQCD")->second << " index = " << index <<  endl; 
     return index; 
}

int TMD::TMDgetOrderPDF( ){
/** TMDlib helper to return order of evolution (nr of loops) from info file */

     int index =  get_key_val_as_int(TMDdict.find("OrderQCD")->second);
     // cout << " TMDgetOrderPDF: " << TMDdict.find("OrderQCD")->second << " index = " << index <<  endl; 
     return index; 

}

double TMD::TMDgetXmin() {
/** TMDlib helper to return XMin of TMD parametrization from info file */
   double XMin =  get_key_val_as_double(TMDdict.find("XMin")->second);
   // cout << " TMDgetXMin as double : " << XMin << endl;
   if(XMin < 0 ) { 
      cout << " TMDgetXMin: XMin is not defined for iset = " << iset << 
      " name = " <<TMDdict.find("XMin")->second << endl; 
      return -9999.; }
   else { 
      return XMin;}

}

double TMD::TMDgetXmax() {
/** TMDlib helper to return XMax of TMD parametrization from info file */
   double XMax =  get_key_val_as_double(TMDdict.find("XMax")->second);
   // cout << " TMDgetXMax as double : " << XMax << endl;
   if(XMax < 0 ) { 
      cout << " TMDgetXMax: XMax is not defined for iset = " << iset << 
      " name = " <<TMDdict.find("XMax")->second << endl; 
      return -9999.; }
   else { 
      return XMax;}
}

double TMD::TMDgetQmin() {
/** TMDlib helper to return QMin of TMD parametrization from info file */
   double QMin =  get_key_val_as_double(TMDdict.find("QMin")->second);
   // cout << " TMDgetQmin as double : " << Qmin << endl;
   if(QMin < 0 ) { 
      cout << " TMDgetQmin: QMin is not defined for iset = " << iset << 
      " value = " <<TMDdict.find("QMin")->second << endl; 
      return -9999.; }
   else { 
      return QMin;}
}

double TMD::TMDgetQ2min() {
/** TMDlib helper to return Q2Min of TMD parametrization from info file */
   double QMin =  get_key_val_as_double(TMDdict.find("QMin")->second);
   // cout << " TMDgetQ2min as double : " << QMin << endl;
   if(QMin < 0 ) { 
      cout << " TMDgetQ2min: Q2Min is not defined for iset = " << iset << " value = " <<TMDdict.find("QMin")->second << endl; 
      return -9999.; }
   else { 
      return QMin*QMin ;}

}
double TMD::TMDgetQ2max() {
/** TMDlib helper to return Q2Max of TMD parametrization from info file */
   double QMax =  get_key_val_as_double(TMDdict.find("QMax")->second);
   // cout << " TMDgetQ2max as double : " << QMax << endl;
   if(QMax < 0 ) { 
      cout << " TMDgetQ2max: QMax is not defined for iset = " << iset << " value = " <<TMDdict.find("QMax")->second <<endl; 
      return -9999.; }
   else { 
      return QMax*QMax; }

}

double TMD::TMDgetQmax() {
/** TMDlib helper to return QMax of TMD parametrization from info file */
   double QMax =  get_key_val_as_double(TMDdict.find("QMax")->second);
   // cout << " TMDgetQmax as double : " << QMax << endl;
   if(QMax < 0 ) { 
      cout << " TMDgetQmax: QMax is not defined for iset = " << iset << " value = " <<TMDdict.find("QMax")->second <<endl; 
      return -9999.; }
   else { 
      return QMax; }

}
string TMD::TMDgetExtrapolation_Q2() {
/** TMDlib helper to return Extrapolation of TMD parametrization from info file */
      return TMDdict.find("Extrapolation_Q2")->second; 

}
string TMD::TMDgetExtrapolation_kt() {
/** TMDlib helper to return Extrapolation of TMD parametrization from info file */
      return TMDdict.find("Extrapolation_x")->second; 

}
string TMD::TMDgetExtrapolation_x() {
/** TMDlib helper to return Extrapolation of TMD parametrization from info file */
      return TMDdict.find("Extrapolation_kt")->second; 

}

int TMD::TMDnumberPDF(string name) {
/** TMDlib helper to return index number for TMDset name from info file */

//   cout << " TMDnumberPDF: " << name <<  endl; 

if(name == "" ) { cout << " TMDnumberPDF: name is not defined " << name << endl; return -99999; }

   else { 
     int index =  get_key_val_as_int(TMDdict.find("SetIndex")->second);
     // cout << " TMDnumberPDF: " << name << " index = " << index <<  endl; 
     return index; }
}

string TMD::TMDstringPDF(int index) {
/** TMDlib helper to return name for index  for TMDset */
     string error="error";
    //  cout << " TMDstringPDF: " << index <<  endl; 

     if(index == 0 ) { cout << " TMDstringPDF: iset is not defined " << index << endl; return error; }

        else { 
       //    static map<int, string> TMDindex;
          if (TMDindex.empty()) { // The map needs to be populated first
              string indexpath =pdfpath+"tmdpdfsets.index";
              // cout << " TMDstringPDF indexpath = " << indexpath << endl;
              if (indexpath.empty()) {cout << "Could not find a tmdpdfsets.index file" <<endl; };
              try {
                 ifstream file(indexpath.c_str());
                 string line;
                 while (getline(file, line)) {
                    if (line.empty() || line.find("#") == 0) continue;
                    istringstream tokens(line);
                    int id; string setname;
                    tokens >> id;
                    tokens >> setname;
                    TMDindex[id] = setname;
                    // cout << id << " -> " << TMDindex[id] << endl;
                 }
             } catch (const std::exception& ex) {
                cout << "Trouble when reading " + indexpath + ": " + ex.what() << endl;
             }
          }
  
         string name = TMDindex.find(index)->second;
         // cout << " TMDstringPDF: " << TMDindex.find(index)->second << " index = " << index <<  endl; 
         return name; }
}
